import importlib.util
import re
from hashlib import sha256
from importlib.machinery import ModuleSpec
from os.path import splitext
from pathlib import Path
from types import ModuleType
from typing import Optional, Union


class BaseHash:

    @staticmethod
    def file(path_file: str):
        """
        Получить хеш сумму данных в файле, по его пути

        :param path_file: Путь к файлу
        """
        h = sha256()
        b = bytearray(128 * 1024)
        mv = memoryview(b)
        with open(path_file, 'rb', buffering=0) as f:
            for n in iter(lambda: f.readinto(mv), 0):
                h.update(mv[:n])
        return h.hexdigest()

    @staticmethod
    def text(text: str) -> str:
        """
        Получить хеш сумму текста
        """
        return sha256(text.encode()).hexdigest()

    @staticmethod
    def check_hash_sum(unknown_hash_sum: str, true_hash_sum: str):
        """
        Сравить хеш суммы

        :param unknown_hash_sum: Полученная(неизвестная) хеш сумма
        :param true_hash_sum: Требуемая хеш сумма
        """
        if unknown_hash_sum != true_hash_sum:
            raise ValueError(f"{unknown_hash_sum} != {true_hash_sum}")
        return True


def read_file_by_module(_path: str) -> ModuleType:
    """
    Импортировать файл как модуль `python`

    :param _path: Путь к `python` файлу
    :return: Модуль `python`
    """
    # Если не нужно проверять имя расширения
    if splitext(_path)[1] != ".py":  # Проверяем расширение файла
        raise ValueError(f"Файл должен иметь расширение .py")
    # указать модуль, который должен быть импортируется относительно пути модуль
    spec: Optional[ModuleSpec] = importlib.util.spec_from_file_location("my_module", _path)
    # создает новый модуль на основе спецификации
    __module: ModuleType = importlib.util.module_from_spec(spec)
    # выполняет модуль в своем собственном пространстве имен,
    # когда модуль импортируется или перезагружается.
    spec.loader.exec_module(__module)
    return __module


def concat_absolute_dir_path(_file: str, _path: str) -> str:
    """
    Получить абсолютный путь папки и объединить с другим путем

    :param _file:
    :param _path:
    :return:
    """
    return str(Path(_file).resolve().parent / _path)


def absolute_path_dir(_file: str, back: int = 1) -> Path:
    """
    Получить абсолютный путь к своей директории

    :param _file: Путь
    :param back: Сколько отступить назад
    """
    res = Path(_file).resolve()
    for _ in range(back):
        res = res.parent
    return res


def toBitSize(size: Union[str, int]) -> int:
    """
    :param size:
    можно указать:
        - kb - Например 10kb
        - mb - Например 1mb

    >>> toBitSize("10kb")
    10240
    >>> toBitSize("1mb")
    1048576
    >>> toBitSize("1gb")
    Traceback (most recent call last):
    ValueError: Не верный тип 1gb
    """
    match size:
        case int():
            return size
        case str() as _res if _r := re.match("([\d_]+)kb|KB", _res):
            return int(_r.group(1)) * 1024
        case str() as _res if _r := re.match("([\d_]+)mb|MB", _res):
            return int(_r.group(1)) * 1048576
        case _:
            raise ValueError(f"Не верный тип {size}")


class T_CryptoAes:
    """
    Источник кода

    https://github.com/denisxab/mg_crp.git
    """

    def __init__(self, key: str): ...

    def encodeAES(self): ...

    def dencodeAES(self): ...

    def __call__(self, *args, **kwargs): ...



from abc import abstractmethod
from os import makedirs, remove, mkdir
from os.path import abspath, dirname, exists, getsize, splitext
from pickle import load, dump
from shutil import rmtree
from typing import Any, Callable, TypeAlias, Union, Optional, NamedTuple

from .helpful import T_CryptoAes, BaseHash


class BaseFile:
    """
    Базовый класс для файлов
    """
    __slots__ = "name_file"

    def __init__(self, name_file: str, type_file: Optional[str]):
        """
        :param type_file: Какое расширение должен иметь файл. Если указать
            None то это значит что проверка попуститься
        :param name_file: Путь к файлу
        """
        self.check_extensions_file(name_file, type_file)
        self.name_file: str = name_file  #: Путь к файлу
        self.createFileIfDoesntExist()

    def createFileIfDoesntExist(self):
        """
        Создать файл если его нет
        """
        if not exists(self.name_file):
            tmp_ = dirname(self.name_file)
            if tmp_ and not exists(tmp_):  # Если задан путь из папок если их нет
                makedirs(tmp_)  # Создаем путь из папок
                open(self.name_file, "w").close()
            else:
                # Если указано только имя файла без папок
                # или папки же существуют
                open(self.name_file, "w").close()

    def checkExistenceFile(self) -> bool:
        """
        Проверить существование файла
        """
        return True if exists(self.name_file) else False

    def deleteFile(self):
        """
        Удалить файл
        """
        # Удаление файла
        if self.checkExistenceFile():
            remove(self.route())

    def sizeFile(self) -> int:
        """
        Получить размер файла
        """
        # Размер файла в байтах
        return getsize(self.name_file)

    def route(self) -> str:  # +
        """
        Получить абсолютный путь
        """
        # Путь к файлу
        return abspath(self.name_file)

    def createRoute(self):
        """
        Создать файл по указному пути, если нужно поместить в папки, то они создадутся
        """
        tmp_route: str = ""
        for folder_name in self.name_file.split('/')[:-1]:
            tmp_route += folder_name
            mkdir(tmp_route)
            tmp_route += '/'

    def removeRoute(self):
        """
        Удалить весь путь к файлу
        """
        rmtree(self.name_file.split('/')[1])

    def encryptFile(self, key: str, CryptoAes: T_CryptoAes, outpath: Optional[str] = str, ):
        """
        Зашифровать файл

        :param key: Ключ для шифрования, должен иметь длину 16,24,32 байта
        :param outpath: Путь куда сохранить зашифрованный файл
        :param CryptoAes: https://github.com/denisxab/mg_crp.git
        """
        res: NamedTuple = CryptoAes(key).encodeAES(str(self.readFile()))
        with open(outpath if outpath else self.name_file, "wb") as _pickFile:
            dump(res, _pickFile, protocol=3)

    def decryptoFile(self, key: str, CryptoAes: T_CryptoAes) -> Optional[str]:
        """
        Расшифровать файл

        :param key: Ключ для шифрования, должен иметь длину 16,24,32 байта
        :param CryptoAes: https://github.com/denisxab/mg_crp.git
        """
        with open(self.name_file, "rb") as _pickFile:
            return CryptoAes(key).decodeAES(load(_pickFile))

    def hashFileSha256(self) -> str:
        """
        Получить хеш сумму файла
        """
        return BaseHash.file(self.name_file)

    @staticmethod
    def check_extensions_file(name_file: str, req_type: str):
        """
        Сравнить расширение файла, с требуемым

        :param name_file: Путь к файлу
        :param req_type: Требуемое расширение
        """
        # Если не нужно проверять имя расширения
        if not req_type:
            return
        if splitext(name_file)[1] != req_type:  # Проверяем расширение файла
            raise ValueError(f"Файл должен иметь расширение {req_type}")

    @abstractmethod
    def readFile(self, *arg) -> Any:
        """
        Прочитать файл
        """
        ...

    @abstractmethod
    def writeFile(self, arg: Any):
        """
       Записать данные в файл
       """
        ...

    @abstractmethod
    def appendFile(self, arg: Any):
        """
        Добавить данные в файл
        """
        ...


T_ConcatData: TypeAlias = Union[list[Union[str, int, float]], list[list[Union[str, int, float]]], dict, set, str]


def concat_data(callback: Callable, file_data: T_ConcatData, new_data: T_ConcatData):
    """
    Объединить два переменных, если они одинакового типа

    :param callback: Вызовется при успешной проверки типов
    :param file_data: Данные из файле
    :param new_data: Текущие данные в `Python`
    """

    # Входные данные должны быть такого же типа, что и в файле
    if type(file_data) == type(new_data):
        match new_data:
            case list():
                file_data.extend(new_data)
            case tuple() | str():
                file_data += new_data
            case dict() | set():
                file_data.update(new_data)
            case _:
                raise TypeError("Не поддерживаемый тип")
        callback(file_data)
    else:
        raise TypeError("Тип данных в файле и тип входных данных различны")



__all__ = ["PickleFile"]

from pickle import load, dump
from typing import Any, Union, Optional

from .base_file import BaseFile, concat_data


class PickleFile(BaseFile):
    """
    Работа с сериализоваными данными


    `Документация pickle
    <https://docs.python.org/3/library/pickle.html>`_

    """

    def __init__(self, name_file: str, type_file: Optional[str] = ".pkl"):
        super().__init__(name_file, type_file=type_file)

    def writeFile(self, data: Any, *, protocol: int = 3):
        """Сериализовать и записать данные в файл"""
        with open(self.name_file, "wb") as _pickFile:
            dump(data, _pickFile, protocol=protocol)

    def readFile(self, **kwargs) -> Any:
        """Прочитать и десериализация данные из файла"""
        with open(self.name_file, "rb") as _pickFile:
            return load(_pickFile)

    def appendFile(self, data: Union[tuple, list, dict, set], *, protocol: int = 3):
        concat_data(
            lambda _data: self.writeFile(_data, protocol=protocol),
            self.readFile(),
            data)



from subprocess import check_output, CalledProcessError, STDOUT
from threading import Lock, Thread

from logsmal import loglevel

# poetry add tqdm
from tqdm import tqdm


def os_exe_thread(
        label_command: str,
        command_list: list[str],
        call_log_info: loglevel = lambda _x, flag: ...,
        call_log_error: loglevel = lambda _x, flag: ...,
):
    """
    Выполнить команды системы в нескольких потоках.


    :param label_command: Общее название команд
    :param command_list: Список команд
    :param call_log_info: Функция для логов информации
    :param call_log_error: Функция для логов ошибок

    :Пример вызова:

    ..code-bloc:: python

        indir = os.path.dirname(__file__)
        command_list: list[str] = []
        command = "pull"

        for _path in listdir():
            command_list.append(f"cd {path.join(indir, _path)} && git {command}")

        os_exe_thread(
            "GIT PULL",
            command_list,
            call_log_info=logger.info,
            call_log_error=logger.error
        )
    """

    lock = Lock()

    def self_(_command: str):
        """

        :param _command:
        """

        try:
            res = check_output(_command, shell=True, stderr=STDOUT)
            with lock:
                call_log_info(f"{_command}:{res.decode('utf-8')}", flag=label_command)
        except CalledProcessError as e:
            with lock:
                call_log_error(f"{_command}:{e.output.decode('utf-8')}", flag=str(e.returncode))
        finally:
            with lock:
                pbar.set_description(f"{_command}")
                pbar.update()

    list_thread: list[Thread] = []
    with tqdm(total=len(command_list)) as pbar:
        for _command in command_list:
            th = Thread(
                target=self_, args=(_command,),
                name=f"th_{_command}", daemon=True,
            )
            list_thread.append(th)
            th.start()

        for th in list_thread:
            th.join()



from mg_crp.crypto_aes import CryptoAes

from mg_file import TxtFile

file_name = "./test.txt"
file_name_out = "./test_crp.txt"
secret_key = "Sixteen byte kys"
data = "Привет мир ВОт мои dAta 123"


class Test_BaseFile:

    def setup(self):  # Выполнятся перед вызовом каждого метода
        ...

    def test_encryptFile(self):  # Все методы должны начинаться со слова `test_`
        _f = TxtFile(file_name)
        _f.writeFile(data)
        _f.encryptFile(secret_key, CryptoAes, outpath=file_name_out)
        res = TxtFile(file_name_out).decryptoFile(secret_key, CryptoAes)
        assert res == data
        _f.deleteFile()
        TxtFile(file_name_out).deleteFile()

    def teardown(self):  # Выполнятся после **успешного** выполнения каждого теста
        ...

    def __del__(self):  # Деструктор класса
        ...



from typing import Optional, Any, Literal

from .txt_file import TxtFile


class LogFile(TxtFile):
    """
    Класс дял работы с лог файлами
    """

    def __init__(self, name_file: str, *, mod: Optional[Literal['r', 'w', 'rb', 'wb', 'a', 'ab']] = None,
                 encoding: str = None, data: Any = None, type_file: Optional[str] = ".log"):
        super().__init__(name_file, mod=mod, encoding=encoding, data=data, type_file=type_file)



import doctest
import unittest

from mg_file.file import helpful


def load_tests(loader, tests, ignore):
    """
    Док тест
    """
    tests.addTests(doctest.DocTestSuite(helpful))
    return tests


if __name__ == '__main__':
    unittest.main()



__all__ = ["TxtFile"]

from typing import Dict, Union, Any, Literal, Optional

from .base_file import BaseFile


class TxtFile(BaseFile):
    """
    Работа с текстовым файлом

    :Example:

    .. code-block:: python

        from  mg_file.file.txt_file import TxtFile
        txt_obj = TxtFile('./path/file.txt')
    """

    def __init__(self, name_file: str, *,
                 mod: Optional[Literal['r', 'w', 'rb', 'wb', 'a', 'ab']] = None,
                 encoding: str = None,
                 data: Any = None,
                 type_file: Optional[str] = ".txt"):
        """
        Вы можете сразу выполнить метод указав ``mod``

        :param name_file:
        :param mod:
        :param encoding:
        :param data:
        :param type_file: Какое расширение должен иметь файл
        """
        super().__init__(name_file, type_file)
        if mod:
            self.res: Union[str, Any] = {
                "r": lambda: self.readFile(encoding=encoding),
                "w": lambda: self.writeFile(data=data),
                "rb": lambda: self.readBinaryFile(),
                "wb": lambda: self.writeBinaryFile(data=data),
                "a": lambda: self.appendFile(data=data),
                "ab": lambda: self.appendBinaryFile(data=data)
            }[mod]()

    def readFile(self, limit: int = 0, *, encoding: str = None) -> str:
        """
        Прочитать файл с начало

        :param limit: Ограничение чтения строк
        :param encoding: Кодировка
        """
        with open(self.name_file, "r", encoding=encoding) as f:
            if limit:
                res: str = ""
                for line in f:
                    if limit:
                        res += line
                        limit -= 1
                    else:
                        break
                return res
            else:
                return f.read()

    def readFileToResDict(self, *args: str, separator: str = '\n') -> Dict[str, str]:
        """
        Прочитать файл и вернуть словарь

        :param separator: Разделитель
        :param args: Имя ключей словаря

        :Пример:

        Файл ``./path/file.txt``

        .. code-block:: txt

            denisxab
            denis-k@mail.com
            password123

        Код

        .. code-block:: python

            from  mg_file.file.txt_file import TxtFile
            TxtFile('./path/file.txt').readFileToResDict("user_name","email","password")
            # {'user_name': 'denisxab', 'email': 'denis-k@mail.com', 'password': 'password123'}
        """
        resDict: Dict[str, str] = {}
        with open(self.name_file, "r") as f:
            for index, line in enumerate(f):
                resDict[args[index]] = line.replace(separator, "")
        return resDict

    def search(self, name_find: str) -> bool:
        """
        Простой поиск на соответствие тексту в файле

        :param name_find: Что искать

        :Пример:

        Файл ``./path/file.txt``

        .. code-block:: txt

            Optional. If the number of
            bytes returned exceed the hint number,
            no more lines will be returned. Default value is  -1,
            which means all lines will be returned.

        Код

        .. code-block:: python

            from  mg_file.file.txt_file import TxtFile
            TxtFile('./path/file.txt').searchFile("Default")
            # True
            TxtFile('./path/file.txt').searchFile("БУКВА")
            # False
        """
        res: bool = False
        with open(self.name_file, "r") as f:
            for line in f:
                if line.find(name_find) != -1:
                    res = True
                    break
        return res

    def readBinaryFile(self) -> bytes:  # +
        """
        Прочитать файл в бинарном режиме
        """
        with open(self.name_file, "rb") as f:
            return f.read()

    def writeFile(self, data: str):
        with open(self.name_file, "w") as f:
            f.write(data)

    def writeBinaryFile(self, data: Union[bytes, memoryview]):
        """
        Записать данные в бинарном режиме
        """
        with open(self.name_file, "wb") as f:
            f.write(data)

    def appendFile(self, data: str):  # +
        with open(self.name_file, "a") as f:
            f.write(data)

    def appendBinaryFile(self, data: bytes):
        """
        Добавить данные в бинарном режиме
        """
        with open(self.name_file, "ab") as f:
            f.write(data)



from .file import *
from .pcos import *
from .sqllite_orm import *
from .sql_raw import *
from .castom_type import *

# from .pcos import *
# from .file import *
# from .sqllite_orm import *
# from .logsmal import *
# from .sql_raw import *
# from .castom_type import *